/*   **********************************************************************  **
 **   Copyright notice                                                       **
 **                                                                          **
 **   (c) 2009, Around(J2)ME				                                 **
 **   All rights reserved.                                                   **
 **                                                                          **
 **   This program and the accompanying materials are made available under   **
 **   the terms of the New BSD License which accompanies this    			 **
 **   distribution.											                 **
 **   A copy is found in the textfile LICENSE.txt							 **
 **                                                                          **
 **   This copyright notice MUST APPEAR in all copies of the file!           **
 **                                                                          **
 **   Main developers:                                                       **
 **     Juri Strumpflohner		http://blog.js-development.com	             **
 **		Matthias Braunhofer		http://matthias.jimdo.com                    **
 **                                                                          **
 **  **********************************************************************  */

package com.aroundj2me.location;

import java.io.IOException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import javax.microedition.location.AddressInfo;
import javax.microedition.location.Coordinates;
import javax.microedition.location.Landmark;
import javax.microedition.location.LandmarkException;
import javax.microedition.location.LandmarkStore;
import javax.microedition.location.Location;
import javax.microedition.location.LocationException;
import javax.microedition.location.LocationProvider;

import com.aroundj2me.model.ContinentName;
import com.aroundj2me.model.Place;
import com.aroundj2me.utility.StringUtils;

public class LocationService {
	//the timeout to determine the current location
	public static final int LOCATION_SEARCH_TIMEOUT = 15;
	//stores the LandmarkStores for each continent such that they can be accessed quickly
	private Hashtable landmarksForContinents = new Hashtable();
	//prefix that is appended to the name of the LandmarkStores
	private String prefix;
	//set radius to approximately 3 km -> 1 ~ 110 km
	private final double RADIUS = 0.03;
	
	public LocationService(String prefix) {
		this.prefix = prefix;	
	}
	
	/**
	 * Initializes the LandmarkStores for all continents
	 * @throws NullPointerException
	 */
	public void init() throws NullPointerException {
		//ASIA
		String storeNameForContinent = prefix + ContinentName.ASIA;
		LandmarkStore store = LandmarkStore.getInstance(storeNameForContinent);
		if (store != null) landmarksForContinents.put(ContinentName.ASIA, store);
		
		//AFRICA
		storeNameForContinent = prefix + ContinentName.AFRICA;
		store = LandmarkStore.getInstance(storeNameForContinent);
		if (store != null) landmarksForContinents.put(ContinentName.AFRICA, store);
		
		//EUROPE
		storeNameForContinent = prefix + ContinentName.EUROPE;
		store = LandmarkStore.getInstance(storeNameForContinent);
		if (store != null) landmarksForContinents.put(ContinentName.EUROPE, store);
		
		//NORTH AMERICA
		storeNameForContinent = prefix + ContinentName.NORTHAMERICA;
		store = LandmarkStore.getInstance(storeNameForContinent);
		if (store != null) landmarksForContinents.put(ContinentName.NORTHAMERICA, store);
		
		//SOUTH AMERICA
		storeNameForContinent = prefix + ContinentName.SOUTHAMERICA;
		store = LandmarkStore.getInstance(storeNameForContinent);
		if (store != null) landmarksForContinents.put(ContinentName.SOUTHAMERICA, store);
		
		//AUSTRALIA
		storeNameForContinent = prefix + ContinentName.AUSTRALIA;
		store = LandmarkStore.getInstance(storeNameForContinent);
		if (store != null) landmarksForContinents.put(ContinentName.AUSTRALIA, store);
	}
	
	/**
	 * Determines whether a LandmarkStore for a certain country has been created
	 * @param continent
	 * @return
	 */
	public boolean existsLandmarkStore(String continent) {
		LandmarkStore landmarkStore = (LandmarkStore)landmarksForContinents.get(continent);
		if (landmarkStore == null) {
			return false;
		}
		return true;
	}
	
	/**
	 * Tries to create a LandmarkStore for a certain continent
	 * @param continent
	 * @throws IOException
	 * @throws LandmarkException
	 */
	public void createLandmarkStore(String continent) throws IOException, LandmarkException {
		if(existsLandmarkStore(continent)){
			deleteLandmarkStore(continent);
		}
		
		//create a new landmark store
		String storeNameForContinent = prefix + continent;
		LandmarkStore.createLandmarkStore(storeNameForContinent);
		landmarksForContinents.put(continent, LandmarkStore.getInstance(storeNameForContinent));
	}

	/**
	 * Adds a category to the LandmarkStore for a certain continent
	 * @param category
	 * @throws LandmarkException 
	 * @throws IOException 
	 * @throws IOException
	 * @throws LandmarkException
	 */
	private void addCategory(String continent, String category) throws IOException, LandmarkException {
		((LandmarkStore)landmarksForContinents.get(continent)).addCategory(category);
	}
	
	/**
	 * Adds a landmark to the LandmarkStore for a certain continent
	 * @param continent
	 * @param landmark
	 * @param categories
	 * @throws IOException
	 * @throws LandmarkException 
	 */
	public void addLandmark(String continent, Landmark landmark, String[] categories) throws IOException, LandmarkException {
				
		int noOfCategories = categories.length;
		for(int i = 0; i < noOfCategories; i++) {
			String category = categories[i];
			if (!existsCategory(continent, category)) {
				addCategory(continent, category);
			}
			((LandmarkStore)landmarksForContinents.get(continent)).addLandmark(landmark, category);	
		}
	}
	
	/**
	 * Retrieves all nearby places for a certain category
	 * @param coordinates
	 * @param category
	 * @return
	 * @throws IOException
	 */
	public Vector findNearbyLandmarks(Coordinates coordinates, String category) throws IOException {
		Vector result = new Vector();
		Place place = null;
		Landmark landmark = null;
		Enumeration storesEnum = landmarksForContinents.elements();
		while (storesEnum != null && storesEnum.hasMoreElements()) {
			LandmarkStore store = (LandmarkStore)storesEnum.nextElement();
			Enumeration landmarksEnum = store.getLandmarks(category, coordinates.getLatitude() - RADIUS, coordinates.getLatitude() + RADIUS, coordinates.getLongitude() - RADIUS, coordinates.getLongitude() + RADIUS);
			while (landmarksEnum != null && landmarksEnum.hasMoreElements()) {
	        	landmark = (Landmark)landmarksEnum.nextElement();
	        	place = new Place();
	        	place.setLandmark(landmark);
	        	place.setDistance(landmark.getQualifiedCoordinates().distance(coordinates));
	        	place.setCategories(StringUtils.split(landmark.getDescription(), ';'));
	        	result.addElement(place);
	        }
		}
		return result;
	}
	
	/**
	 * Retrieves all nearby places that match a certain search term
	 * @param coordinates
	 * @param searchTerm
	 * @return
	 * @throws IOException
	 */
	public Vector findNearbyLandmarksByTerm(Coordinates coordinates, String searchTerm) throws IOException {
		searchTerm = searchTerm.toLowerCase();
		Vector places = findNearbyLandmarks(coordinates);
		int noOfPlaces = places.size();
		for (int i = noOfPlaces - 1; i >= 0; i--) {
			Place place = (Place)places.elementAt(i);
			Landmark landmark = place.getLandmark();
			AddressInfo addressInfo = landmark.getAddressInfo();
			if (landmark.getName().toLowerCase().indexOf(searchTerm) == -1 && addressInfo.getField(AddressInfo.STREET).toLowerCase().indexOf(searchTerm) == -1 && 
				addressInfo.getField(AddressInfo.URL).toLowerCase().indexOf(searchTerm) == -1) {
				String[] categories = place.getCategories();
				int noOfCategories = categories.length;
				boolean matchesCategory = false;
				for (int j = 0; j < noOfCategories; j++) {
					String category = categories[j];
					if (category.toLowerCase().indexOf(searchTerm) != -1) {
						matchesCategory = true;
						break;
					}
				}
				if (!matchesCategory) {
					places.removeElement(place);
				}	
			}
		}
		return places;
	}
	
	/**
	 * Retrieves all nearby landmarks
	 * @param coordinates
	 * @return
	 * @throws IOException 
	 * @throws IOException
	 */
	public Vector findNearbyLandmarks(Coordinates coordinates) throws IOException {
		return findNearbyLandmarks(coordinates, null);
	}
	
	/**
	 * Deletes the complete LandmarkStore for a continent
	 * @param continent
	 * @throws IOException
	 * @throws LandmarkException
	 */
	private void deleteLandmarkStore(String continent) throws IOException, LandmarkException {
		String recordStoreName = prefix + continent;
		LandmarkStore.deleteLandmarkStore(recordStoreName);
		landmarksForContinents.remove(continent);
	}
	
	/**
	 * Checks whether the passed category already exists in the LandmarkStore of a certain continent
	 * @param continent
	 * @param category
	 * @return
	 */
	private boolean existsCategory(String continent, String category) {
		Enumeration categoriesEnum = ((LandmarkStore)landmarksForContinents.get(continent)).getCategories();
		while (categoriesEnum.hasMoreElements()) {
			String cat = (String)categoriesEnum.nextElement();
			if (category.equals(cat)) {
				return true;
			}
		}
		return false;
	}
	
	/**
	 * Retrieves the current user location. Attention, this has to be invoked from within a separate thread!
	 * @param timeout
	 * @return
	 * @throws LocationException
	 * @throws InterruptedException
	 */
	public Location getCurrentLocation() throws LocationException, InterruptedException{
		LocationProvider provider = LocationProvider.getInstance(null);
		return provider.getLocation(LOCATION_SEARCH_TIMEOUT);
	}
}
